from typing import TYPE_CHECKING, Optional

import attr

from dis_snek.const import MISSING
from dis_snek.models.discord import DiscordObject
from dis_snek.models.snowflake import to_snowflake
from dis_snek.models.enums import StagePrivacyLevel
from dis_snek.utils.attr_utils import define

if TYPE_CHECKING:
    from dis_snek.models import Guild, GuildStageVoice, Snowflake_Type


@define
class StageInstance(DiscordObject):
    topic: str = attr.ib()
    privacy_level: StagePrivacyLevel = attr.ib()
    discoverable_disabled: bool = attr.ib()

    _guild_id: "Snowflake_Type" = attr.ib(converter=to_snowflake)
    _channel_id: "Snowflake_Type" = attr.ib(converter=to_snowflake)

    @property
    def guild(self) -> "Guild":
        return self._client.cache.guild_cache.get(self._guild_id)

    @property
    def channel(self) -> "GuildStageVoice":
        return self._client.cache.channel_cache.get(self._channel_id)

    async def delete(self, reason: Optional[str] = MISSING):
        """
        Delete this stage instance. Effectively closes the stage

        Args:
            reason: The reason for this deletion, for the audit log
        """
        await self._client.http.delete_stage_instance(self._channel_id, reason)



from typing import Optional, TYPE_CHECKING, Union
import attr

if TYPE_CHECKING:
    from io import IOBase
    from pathlib import Path


@attr.s()
class File:
    """Representation of a file. Used to"""

    file: Union["IOBase", "Path", str] = attr.field()
    """Location of file to send or the bytes."""
    file_name: Optional[str] = attr.field(default=None)
    """Set a filename that will be displayed when uploaded to discord. If you leave this empty, the file will be called `file` by default"""



import copy
import logging

from dis_snek.const import logger_name, MISSING
from dis_snek.event_processors._template import EventMixinTemplate
from dis_snek.models import listen, events, GuildIntegration
from dis_snek.models.events import RawGatewayEvent
from dis_snek.models.events.discord import IntegrationCreate, IntegrationDelete, BanCreate, BanRemove

log = logging.getLogger(logger_name)


class GuildEvents(EventMixinTemplate):
    @listen()
    async def _on_raw_guild_create(self, event: RawGatewayEvent) -> None:
        """
        Automatically cache a guild upon GUILD_CREATE event from gateway.

        Args:
            event: raw guild create event
        """
        guild = self.cache.place_guild_data(event.data)
        self._guild_event.set()

        self.dispatch(events.GuildJoin(guild))

    @listen()
    async def _on_raw_guild_update(self, event: RawGatewayEvent) -> None:
        before = copy.copy(await self.cache.get_guild(event.data.get("id")))
        self.dispatch(events.GuildUpdate(before or MISSING, self.cache.place_guild_data(event.data)))

    @listen()
    async def _on_raw_guild_delete(self, event: RawGatewayEvent) -> None:
        if event.data.get("unavailable", False):
            self.dispatch(
                events.GuildUnavailable(
                    event.data.get("id"),
                    await self.cache.get_guild(event.data.get("id"), False) or MISSING,
                )
            )
        else:
            self.dispatch(
                events.GuildLeft(
                    event.data.get("id"),
                    await self.cache.get_guild(event.data.get("id"), False) or MISSING,
                )
            )

    @listen()
    async def _on_raw_guild_ban_add(self, event: RawGatewayEvent) -> None:
        self.dispatch(BanCreate(event.data.get("guild_id"), self.cache.place_user_data(event.data.get("user"))))

    @listen()
    async def _on_raw_guild_ban_remove(self, event: RawGatewayEvent) -> None:
        self.dispatch(BanRemove(event.data.get("guild_id"), self.cache.place_user_data(event.data.get("user"))))

    @listen()
    async def _on_raw_integration_create(self, event: RawGatewayEvent) -> None:
        self.dispatch(IntegrationCreate(GuildIntegration.from_dict(event.data, self)))  # type: ignore

    @listen()
    async def _on_raw_integration_update(self, event: RawGatewayEvent) -> None:
        self.dispatch(IntegrationUpdate(GuildIntegration.from_dict(event.data, self)))  # type: ignore

    @listen()
    async def _on_raw_integration_delete(self, event: RawGatewayEvent) -> None:
        self.dispatch(
            IntegrationDelete(event.data.get("guild_id"), event.data.get("id"), event.data.get("application_id"))
        )



import asyncio
import inspect
import logging
import re
from enum import IntEnum
from typing import TYPE_CHECKING, Callable, Coroutine, Dict, List, Union, Optional, Any

import attr

from dis_snek.const import (
    GLOBAL_SCOPE,
    CONTEXT_MENU_NAME_LENGTH,
    SLASH_OPTION_NAME_LENGTH,
    SLASH_CMD_NAME_LENGTH,
    SLASH_CMD_MAX_OPTIONS,
    SLASH_CMD_MAX_DESC_LENGTH,
    MISSING,
    logger_name,
)
from dis_snek.mixins.serialization import DictSerializationMixin
from dis_snek.models.command import BaseCommand
from dis_snek.models.discord_objects.channel import BaseChannel
from dis_snek.models.discord_objects.role import Role
from dis_snek.models.discord_objects.user import BaseUser
from dis_snek.models.enums import ChannelTypes, CommandTypes
from dis_snek.models.snowflake import to_snowflake, to_snowflake_list
from dis_snek.utils.attr_utils import docs
from dis_snek.utils.misc_utils import get_parameters
from dis_snek.utils.serializer import no_export_meta

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type

log = logging.getLogger(logger_name)


class OptionTypes(IntEnum):
    """Option types supported by slash commands."""

    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10

    @classmethod
    def from_type(cls, t: type) -> "OptionTypes":
        """
        Convert data types to their corresponding OptionType.

        parameters:
            t: The datatype to convert

        returns:
            OptionType or None
        """
        if issubclass(t, str):
            return cls.STRING
        if issubclass(t, int):
            return cls.INTEGER
        if issubclass(t, bool):
            return cls.BOOLEAN
        if issubclass(t, BaseUser):
            return cls.USER
        if issubclass(t, BaseChannel):
            return cls.CHANNEL
        if issubclass(t, Role):
            return cls.ROLE
        if issubclass(t, float):
            return cls.NUMBER


class PermissionTypes(IntEnum):
    """Types of target supported by the interaction permission."""

    ROLE = 1
    USER = 2

    @classmethod
    def from_type(cls, t: type) -> "PermissionTypes":
        if issubclass(t, Role):
            return cls.ROLE
        if issubclass(t, BaseUser):
            return cls.USER


class CallbackTypes(IntEnum):
    """Types of callback supported by interaction response."""

    PONG = 1
    CHANNEL_MESSAGE_WITH_SOURCE = 4
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7
    AUTOCOMPLETE_RESULT = 8


@attr.s(slots=True)
class Permission:
    """
    Represents a interaction permission.

    parameters:
        id: The id of the role or user.
        guild_id: The guild this permission belongs to
        type: The type of id (user or role)
        permission: The state of permission. ``True`` to allow, ``False``, to disallow.
    """

    id: "Snowflake_Type" = attr.ib(converter=to_snowflake)
    guild_id: "Snowflake_Type" = attr.ib(converter=to_snowflake, metadata=no_export_meta)
    type: Union[PermissionTypes, int] = attr.ib(converter=PermissionTypes)
    permission: bool = attr.ib(default=True)

    def to_dict(self) -> dict:
        """
        Convert this object into a dict ready for discord.

        returns:
            Representation of this object
        """
        data = attr.asdict(self)
        data.pop("guild_id", None)
        return data


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class InteractionCommand(BaseCommand):
    """
    Represents a discord abstract interaction command.

    parameters:
        scope: Denotes whether its global or for specific guild.
        default_permission: Is this command available to all users?
        permissions: Map of guild id and its respective list of permissions to apply.
        cmd_id: The id of this command given by discord.
        callback: The coroutine to callback when this interaction is received.
    """

    name: str = attr.ib(metadata=docs("1-32 character name") | no_export_meta)
    scopes: List["Snowflake_Type"] = attr.ib(
        default=[GLOBAL_SCOPE],
        converter=to_snowflake_list,
        metadata=docs("The scopes of this interaction. Global or guild ids") | no_export_meta,
    )

    default_permission: bool = attr.ib(
        default=True, metadata=docs("whether this command is enabled by default when the app is added to a guild")
    )
    permissions: Optional[List[Union[Permission, Dict]]] = attr.ib(
        factory=dict, metadata=docs("The permissions of this interaction")
    )

    cmd_id: "Snowflake_Type" = attr.ib(
        default=None, metadata=docs("The unique ID of this interaction") | no_export_meta
    )
    callback: Callable[..., Coroutine] = attr.ib(
        default=None, metadata=docs("The coroutine to call when this interaction is received") | no_export_meta
    )

    @property
    def resolved_name(self):
        """A representation of this interaction's name"""
        return self.name

    @property
    def is_subcommand(self) -> bool:
        return False


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class ContextMenu(InteractionCommand):
    """
    Represents a discord context menu.

    parameters:
        name: The name of this entry.
        type: The type of entry (user or message).
    """

    name: str = attr.ib(metadata=docs("1-32 character name"))
    type: CommandTypes = attr.ib(metadata=docs("The type of command, defaults to 1 if not specified"))

    @name.validator
    def _name_validator(self, attribute: str, value: str) -> None:
        if not 1 <= len(value) <= CONTEXT_MENU_NAME_LENGTH:
            raise ValueError("Context Menu name attribute must be between 1 and 32 characters")

    @type.validator
    def _type_validator(self, attribute: str, value: int):
        if not isinstance(value, CommandTypes):
            if value not in CommandTypes.__members__.values():
                raise ValueError("Context Menu type not recognised, please consult the docs.")
        elif value == CommandTypes.CHAT_INPUT:
            raise ValueError(
                "The CHAT_INPUT type is basically slash commands. Please use the @slash_command() " "decorator instead."
            )


@attr.s(slots=True)
class SlashCommandChoice(DictSerializationMixin):
    """
    Represents a discord slash command choice.

    parameters:
        name: The name the user will see
        value: The data sent to your code when this choice is used
    """

    name: str = attr.ib()
    value: Union[str, int, float] = attr.ib()


@attr.s(slots=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class SlashCommandOption(DictSerializationMixin):
    """
    Represents a discord slash command option.

    parameters:
        name: The name of this option
        type: The type of option
        description: The description of this option
        required: "This option must be filled to use the command"
        choices: A list of choices the user has to pick between
        channel_types: The channel types permitted. The option needs to be a channel
        min_value: The minimum value permitted. The option needs to be an integer or float
        max_value: The maximum value permitted. The option needs to be an integer or float
    """

    name: str = attr.ib()
    type: Union[OptionTypes, int] = attr.ib()
    description: str = attr.ib(default="No Description Set")
    required: bool = attr.ib(default=True)
    autocomplete: bool = attr.ib(default=False)
    choices: List[Union[SlashCommandChoice, Dict]] = attr.ib(factory=list)
    channel_types: Optional[list[Union[ChannelTypes, int]]] = attr.ib(default=None)
    min_value: Optional[float] = attr.ib(default=None)
    max_value: Optional[float] = attr.ib(default=None)

    @name.validator
    def _name_validator(self, attribute: str, value: str) -> None:
        if not re.match(rf"^[\w-]{{1,{SLASH_CMD_NAME_LENGTH}}}$", value) or value != value.lower():
            raise ValueError(
                f"Options names must be lower case and match this regex: ^[\w-]{1, {SLASH_CMD_NAME_LENGTH} }$"
            )  # noqa: W605

    @description.validator
    def _description_validator(self, attribute: str, value: str) -> None:
        if not 1 <= len(value) <= SLASH_OPTION_NAME_LENGTH:
            raise ValueError("Options must be between 1 and 100 characters long")

    @type.validator
    def _type_validator(self, attribute: str, value: int) -> None:
        if value == OptionTypes.SUB_COMMAND or value == OptionTypes.SUB_COMMAND_GROUP:
            raise ValueError(
                "Options cannot be SUB_COMMAND or SUB_COMMAND_GROUP. If you want to use subcommands, "
                "see the @sub_command() decorator."
            )

    @channel_types.validator
    def _channel_types_validator(self, attribute: str, value: Optional[list[OptionTypes]]) -> None:
        if value is not None:
            if self.type != OptionTypes.CHANNEL:
                raise ValueError("The option needs to be CHANNEL to use this")

            allowed_int = [channel_type.value for channel_type in ChannelTypes]
            for item in value:
                if (item not in allowed_int) and (item not in ChannelTypes):
                    raise ValueError(f"{value} is not allowed here")

    @min_value.validator
    def _min_value_validator(self, attribute: str, value: Optional[float]) -> None:
        if value is not None:
            if self.type != OptionTypes.INTEGER and self.type != OptionTypes.NUMBER:
                raise ValueError("`min_value` can only be supplied with int or float options")

            if self.type == OptionTypes.INTEGER:
                if isinstance(value, float):
                    raise ValueError("`min_value` needs to be an int in an int option")

            if self.max_value is not None and self.min_value is not None:
                if self.max_value < self.min_value:
                    raise ValueError("`min_value` needs to be <= than `max_value`")

    @max_value.validator
    def _max_value_validator(self, attribute: str, value: Optional[float]) -> None:
        if value is not None:
            if self.type != OptionTypes.INTEGER and self.type != OptionTypes.NUMBER:
                raise ValueError("`max_value` can only be supplied with int or float options")

            if self.type == OptionTypes.INTEGER:
                if isinstance(value, float):
                    raise ValueError("`max_value` needs to be an int in an int option")

            if self.max_value and self.min_value:
                if self.max_value < self.min_value:
                    raise ValueError("`min_value` needs to be <= than `max_value`")


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class SlashCommand(InteractionCommand):
    name: str = attr.ib()
    description: str = attr.ib("No Description Set")

    group_name: str = attr.ib(default=None, metadata=no_export_meta)
    group_description: str = attr.ib(default="No Description Set", metadata=no_export_meta)

    sub_cmd_name: str = attr.ib(default=None, metadata=no_export_meta)
    sub_cmd_description: str = attr.ib(default="No Description Set", metadata=no_export_meta)

    options: List[Union[SlashCommandOption, Dict]] = attr.ib(factory=list)
    autocomplete_callbacks: dict = attr.ib(factory=dict, metadata=no_export_meta)

    @property
    def resolved_name(self):
        return f"{self.name}{f' {self.group_name}' if self.group_name else ''}{f' {self.sub_cmd_name}' if self.sub_cmd_name else ''}"

    @property
    def is_subcommand(self) -> bool:
        return self.sub_cmd_name is not None

    def __attrs_post_init__(self):
        params = get_parameters(self.callback)
        for name, val in params.items():
            if val.annotation and isinstance(val.annotation, SlashCommandOption):

                if not self.options:
                    self.options = []
                val.annotation.name = name
                self.options.append(val.annotation)

        if self.callback is not None:
            if hasattr(self.callback, "options"):
                if not self.options:
                    self.options = []
                self.options += self.callback.options

            if hasattr(self.callback, "permissions"):
                self.permissions = self.callback.permissions
        super().__attrs_post_init__()

    def to_dict(self) -> dict:
        data = super().to_dict()
        if self.is_subcommand:
            data["name"] = self.sub_cmd_name
            data["description"] = self.sub_cmd_description
            data.pop("default_permission", None)
        return data

    @name.validator
    @group_name.validator
    @sub_cmd_name.validator
    def name_validator(self, attribute: str, value: str) -> None:
        if value:
            if not re.match(rf"^[\w-]{{1,{SLASH_CMD_NAME_LENGTH}}}$", value) or value != value.lower():
                raise ValueError(
                    f"Slash Command names must be lower case and match this regex: ^[\w-]{1, {SLASH_CMD_NAME_LENGTH} }$"
                )  # noqa: W605

    @description.validator
    @group_description.validator
    @sub_cmd_description.validator
    def description_validator(self, attribute: str, value: str) -> None:
        if not 1 <= len(value) <= SLASH_CMD_MAX_DESC_LENGTH:
            raise ValueError(f"Description must be between 1 and {SLASH_CMD_MAX_DESC_LENGTH} characters long")

    @options.validator
    def options_validator(self, attribute: str, value: List) -> None:
        if value:
            if isinstance(value, list):
                if len(value) > SLASH_CMD_MAX_OPTIONS:
                    raise ValueError(f"Slash commands can only hold {SLASH_CMD_MAX_OPTIONS} options")
                if value != sorted(
                    value,
                    key=lambda x: x.required if isinstance(x, SlashCommandOption) else x["required"],
                    reverse=True,
                ):
                    raise ValueError("Required options must go before optional options")

            else:
                raise TypeError("Options attribute must be either None or a list of options")

    def autocomplete(self, option_name: str):
        """A decorator to declare a coroutine as an option autocomplete"""

        def wrapper(call: Callable[..., Coroutine]):
            if not asyncio.iscoroutinefunction(call):
                raise TypeError("autocomplete must be coroutine")
            self.autocomplete_callbacks[option_name] = call

            # automatically set the option's autocomplete attribute to True
            for opt in self.options:
                if isinstance(opt, dict) and opt["name"] == option_name:
                    opt["autocomplete"] = True
                elif isinstance(opt, SlashCommandOption) and opt.name == option_name:
                    opt.autocomplete = True

            return call

        option_name = option_name.lower()
        return wrapper

    def subcommand(
        self,
        sub_cmd_name: str,
        group_name: str = None,
        group_description: str = "No Description Set",
        sub_cmd_description: str = "No Description Set",
        options: List[Union[SlashCommandOption, Dict]] = None,
    ) -> Callable[..., "SlashCommand"]:
        def wrapper(call: Callable[..., Coroutine]) -> "SlashCommand":
            if not asyncio.iscoroutinefunction(call):
                raise TypeError("Subcommand must be coroutine")
            return SlashCommand(
                name=self.name,
                description=self.description,
                group_name=group_name,
                group_description=group_description,
                sub_cmd_name=sub_cmd_name,
                sub_cmd_description=sub_cmd_description,
                options=options,
                callback=call,
            )

        return wrapper


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class ComponentCommand(InteractionCommand):
    # right now this adds no extra functionality, but for future dev ive implemented it
    listeners: list[str] = attr.ib(factory=list)


##############
# Decorators #
##############


def slash_command(
    name: str,
    description: str = "No Description Set",
    scopes: List["Snowflake_Type"] = MISSING,
    options: Optional[List[Union[SlashCommandOption, Dict]]] = None,
    default_permission: bool = True,
    permissions: Optional[List[Union[Permission, Dict]]] = None,
    sub_cmd_name: str = None,
    group_name: str = None,
    sub_cmd_description: str = "No Description Set",
    group_description: str = "No Description Set",
):
    """
    A decorator to declare a coroutine as a slash command.

    note:
        While the base and group descriptions arent visible in the discord client, currently.
        We strongly advise defining them anyway, if you're using subcommands, as Discord has said they will be visible in
        one of the future ui updates.

    parameters:
        name: 1-32 character name of the command
        description: 1-100 character description of the command
        scope: The scope this command exists within
        options: The parameters for the command, max 25
        default_permission: Whether the command is enabled by default when the app is added to a guild
        permissions: The roles or users who can use this command
        sub_cmd_name: 1-32 character name of the subcommand
        sub_cmd_description: 1-100 character description of the subcommand
        group_name: 1-32 character name of the group
        group_description: 1-100 character description of the group

    returns:
        SlashCommand Object
    """

    def wrapper(func) -> SlashCommand:
        if not asyncio.iscoroutinefunction(func):
            raise ValueError("Commands must be coroutines")

        cmd = SlashCommand(
            name=name,
            group_name=group_name,
            group_description=group_description,
            sub_cmd_name=sub_cmd_name,
            sub_cmd_description=sub_cmd_description,
            description=description,
            scopes=scopes if scopes else [GLOBAL_SCOPE],
            default_permission=default_permission,
            permissions=permissions or {},
            callback=func,
            options=options,
        )

        return cmd

    return wrapper


def context_menu(
    name: str,
    context_type: "CommandTypes",
    scopes: List["Snowflake_Type"] = MISSING,
    default_permission: bool = True,
    permissions: Optional[List[Union[Permission, Dict]]] = None,
):
    """
    A decorator to declare a coroutine as a Context Menu

    parameters:
        name: 1-32 character name of the context menu
        context_type: The type of context menu
        scope: The scope this command exists within
        default_permission: Whether the menu is enabled by default when the app is added to a guild
        permissions: The roles or users who can use this menu

    returns:
        ContextMenu object
    """

    def wrapper(func) -> ContextMenu:
        if not asyncio.iscoroutinefunction(func):
            raise ValueError("Commands must be coroutines")

        perm = permissions
        if hasattr(func, "permissions"):
            if perm:
                perm.update(func.permissions)
            else:
                perm = func.permissions

        cmd = ContextMenu(
            name=name,
            type=context_type,
            scopes=scopes if scopes else [GLOBAL_SCOPE],
            default_permission=default_permission,
            permissions=perm,
            callback=func,
        )
        return cmd

    return wrapper


def component_callback(*custom_id: str):
    """
    Register a coroutine as a component callback.

    Component callbacks work the same way as commands, just using components as a way of invoking, instead of messages.
    Your callback will be given a single argument, `ComponentContext`

    Args:
        custom_id: The custom ID of the component to wait for
    """

    def wrapper(func) -> ComponentCommand:
        if not asyncio.iscoroutinefunction(func):
            raise ValueError("Commands must be coroutines")

        return ComponentCommand(name=f"ComponentCallback::{custom_id}", callback=func, listeners=custom_id)

    # allows a mixture of generators and strings to be passed
    unpack = []
    for c in custom_id:
        if inspect.isgenerator(c):
            unpack += list(c)
        else:
            unpack.append(c)
    custom_id = unpack
    return wrapper


def slash_option(
    name: str,
    description: str,
    opt_type: Union[OptionTypes, int],
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union[SlashCommandChoice, dict]] = None,
    channel_types: Optional[list[Union[ChannelTypes, int]]] = None,
    min_value: Optional[float] = None,
    max_value: Optional[float] = None,
) -> Any:
    """
    A decorator to add an option to a slash command.

    parameters:
        name: 1-32 lowercase character name matching ^[\w-]{1,32}$
        opt_type: The type of option
        description: 1-100 character description of option
        required: If the parameter is required or optional--default false
        choices: A list of choices the user has to pick between (max 25)
        channel_types: The channel types permitted. The option needs to be a channel
        min_value: The minimum value permitted. The option needs to be an integer or float
        max_value: The maximum value permitted. The option needs to be an integer or float
    """

    def wrapper(func):
        if hasattr(func, "cmd_id"):
            raise Exception("slash_option decorators must be positioned under a slash_command decorator")

        option = SlashCommandOption(
            name=name,
            type=opt_type,
            description=description,
            required=required,
            autocomplete=autocomplete,
            choices=choices if choices else [],
            channel_types=channel_types,
            min_value=min_value,
            max_value=max_value,
        )
        if not hasattr(func, "options"):
            func.options = []
        func.options.insert(0, option)
        return func

    return wrapper


def slash_permission(guild_id: "Snowflake_Type", permissions: List[Union[Permission, Dict]]) -> Any:
    """
    A decorator to add permissions for a guild to a slash command or context menu.

    parameters:
        guild_id: The target guild to apply the permissions.
        permissions: A list of interaction permission rights.
    """
    guild_id = to_snowflake(guild_id)

    def wrapper(func):
        if hasattr(func, "cmd_id"):
            raise Exception("slash_option decorators must be positioned under a slash_command decorator")

        if not hasattr(func, "permissions"):
            func.permissions = {}

        if guild_id not in func.permissions:
            func.permissions[guild_id] = []
        func.permissions[guild_id] += permissions
        return func

    return wrapper


def application_commands_to_dict(commands: Dict["Snowflake_Type", Dict[str, InteractionCommand]]) -> dict:
    """Convert the command list into a format that would be accepted by discord

    `Snake.interactions` should be the variable passed to this"""
    cmd_bases = {}  # {cmd_base: [commands]}
    """A store of commands organised by their base command"""
    output = {}
    """The output dictionary"""

    def squash_subcommand(subcommands: List) -> Dict:
        output_data = {}
        groups = {}
        sub_cmds = []
        for subcommand in subcommands:
            if not output_data:
                output_data = {
                    "name": subcommand.name,
                    "description": subcommand.description,
                    "options": [],
                    "permissions": subcommand.permissions,
                    "default_permission": subcommand.default_permission,
                }
            if subcommand.group_name:
                if subcommand.group_name not in groups:
                    groups[subcommand.group_name] = {
                        "name": subcommand.group_name,
                        "description": subcommand.group_description,
                        "type": int(OptionTypes.SUB_COMMAND_GROUP),
                        "options": [],
                    }
                groups[subcommand.group_name]["options"].append(
                    subcommand.to_dict() | {"type": int(OptionTypes.SUB_COMMAND)}
                )
            else:
                sub_cmds.append(subcommand.to_dict() | {"type": int(OptionTypes.SUB_COMMAND)})
        options = [g for g in groups.values()] + sub_cmds
        output_data["options"] = options
        return output_data

    for scope, cmds in commands.items():
        for cmd in cmds.values():
            if cmd.name not in cmd_bases:
                cmd_bases[cmd.name] = [cmd]
                continue
            if cmd not in cmd_bases[cmd.name]:
                cmd_bases[cmd.name].append(cmd)

    for cmd_list in cmd_bases.values():
        if any(c.is_subcommand for c in cmd_list):
            # validate all commands share required attributes
            scopes: list[Snowflake_Type] = list(set(s for c in cmd_list for s in c.scopes))
            permissions: dict = {k: v for c in cmd_list for k, v in c.permissions.items()}
            base_description = next(
                (c.description for c in cmd_list if c.description is not None), "No Description Set"
            )

            if not all(c.description in (base_description, "No Description Set") for c in cmd_list):
                log.warning(
                    f"Conflicting descriptions found in `{cmd_list[0].name}` subcommands; `{base_description}` will be used"
                )
            if not all(c.default_permission == cmd_list[0].default_permission for c in cmd_list):
                raise ValueError(f"Conflicting `default_permission` values found in `{cmd_list[0].name}`")

            for cmd in cmd_list:
                cmd.scopes = list(scopes)
                cmd.permissions = permissions
                cmd.description = base_description
            # end validation of attributes
            cmd_data = squash_subcommand(cmd_list)
        else:
            scopes = cmd_list[0].scopes
            cmd_data = cmd_list[0].to_dict()
        for s in scopes:
            if s not in output:
                output[s] = [cmd_data]
                continue
            output[s].append(cmd_data)

    return output


def _compare_options(local_opt_list: dict, remote_opt_list: dict):
    if local_opt_list != remote_opt_list:
        if len(local_opt_list) != len(remote_opt_list):
            return False
        for i in range(len(local_opt_list)):
            local_option = local_opt_list[i]
            remote_option = remote_opt_list[i]
            if local_option["type"] == remote_option["type"]:
                if local_option["type"] in (OptionTypes.SUB_COMMAND_GROUP, OptionTypes.SUB_COMMAND):
                    if not _compare_options(local_option["options"], remote_option["options"]):
                        return False
                else:
                    if (
                        local_option["name"] != remote_option["name"]
                        or local_option["description"] != remote_option["description"]
                        or local_option["required"] != remote_option.get("required", False)
                        or local_option["autocomplete"] != remote_option.get("autocomplete", False)
                    ):
                        return False
    return True


def sync_needed(local_cmd: dict, remote_cmd: Optional[dict] = None) -> bool:
    """
    Compares a local application command to its remote counterpart to determine if a sync is required.

    Args:
        local_cmd: The local json representation of the command
        remote_cmd: The json representation of the command from Discord

    Returns:
        Boolean indicating if a sync is needed
    """
    if not remote_cmd:
        # No remote version, command must be new
        return True

    if (
        local_cmd["name"] != remote_cmd["name"]
        or local_cmd.get("description", "") != remote_cmd.get("description", "")
        or local_cmd["default_permission"] != remote_cmd["default_permission"]
    ):
        # basic comparison of attributes
        return True

    if remote_cmd["type"] == CommandTypes.CHAT_INPUT:
        try:
            if not _compare_options(local_cmd["options"], remote_cmd["options"]):
                # options are not the same, sync needed
                return True
        except KeyError:
            if "options" in local_cmd or "options" in remote_cmd:
                return True

    return False



import asyncio
from asyncio import QueueEmpty
from collections import namedtuple
from typing import TYPE_CHECKING

import attr

from dis_snek.const import MISSING
from dis_snek.models.discord import ClientObject
from dis_snek.models.discord_objects.emoji import Emoji
from dis_snek.models.iterator import AsyncIterator
from dis_snek.models.snowflake import to_snowflake
from dis_snek.utils.attr_utils import define

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type
    from dis_snek.models.discord_objects.user import BaseUser


class ReactionUsers(AsyncIterator):
    """
    An async iterator for searching through a channel's history

    Args:
        channel_id: The ID of the channel to search through
        limit: The maximum number of users to return (set to 0 for no limit)
        after: get users after this message ID
    """

    def __init__(self, reaction: "Reaction", limit=50, after=None):
        self.reaction: "Reaction" = reaction
        self.after: "Snowflake_Type" = after
        self._more = True
        super().__init__(limit)

    async def fetch(self):
        if self._more:
            expected = self.get_limit

            if self.after and not self.last:
                self.last = namedtuple("temp", "id")
                self.last.id = self.after

            users = await self.reaction._client.http.get_reactions(
                self.reaction._channel_id,
                self.reaction._message_id,
                self.reaction.emoji.req_format,
                limit=expected,
                after=self.last.id or MISSING,
            )
            if not users:
                raise QueueEmpty()
            self._more = len(users) == expected
            return [self.reaction._client.cache.place_user_data(u) for u in users]
        else:
            raise QueueEmpty()


@define()
class Reaction(ClientObject):
    count: int = attr.ib()
    me: bool = attr.ib(default=False)
    emoji: "Emoji" = attr.ib(converter=Emoji.from_dict)

    _channel_id: "Snowflake_Type" = attr.ib(converter=to_snowflake)
    _message_id: "Snowflake_Type" = attr.ib(converter=to_snowflake)

    def users(self, limit: int = 0, after=None) -> ReactionUsers:
        return ReactionUsers(self, limit, after)

    @property
    def message(self):
        return self._client.cache.message_cache.get((self._channel_id, self._message_id))

    @property
    def channel(self):
        return self._client.cache.channel_cache.get(self._channel_id)

    async def remove(self):
        """Remove all this emoji's reactions from the message"""
        await self._client.http.clear_reaction(self._channel_id, self._message_id, self.emoji.req_format)



from typing import Union, TYPE_CHECKING, Any, List, Type, Dict

from dis_snek.const import MISSING
from dis_snek.utils.misc_utils import get_parameters

from dis_snek.models.context import Context
from dis_snek.models.scale import Scale
from dis_snek.client import Snake

if TYPE_CHECKING:
    from dis_snek.models import Member, User, TYPE_MESSAGEABLE_CHANNEL


def define_annotation():
    """
    Define a function as an annotation.

    The function **must** type hint its expected arguments. Make sure your annotation runs quickly,
    as command execution will be delayed until all annotations are completed

    **Supported Types:**
    `Context`
    `Scale`
    """

    def wrapper(func):
        params = get_parameters(func)
        args = []
        for param in params.values():
            if issubclass(param.annotation, Context):
                args.append("context")
            elif param.annotation is Scale:
                args.append("scale")
            # elif param.annotation is Snake:
            #     args.append("snake")
        func._annotation_dat = {"args": args}
        return func

    return wrapper


@define_annotation()
def CMD_BODY(context: Context) -> str:
    """
    This argument is for the body of the message. IE:

    if `@bot hello how are you?` is sent this argument will be `hello how are you?`
    """
    return context.content_parameters


@define_annotation()
def CMD_AUTHOR(context: Context) -> Union["Member", "User"]:
    """This argument is the author of the command"""

    return context.author


@define_annotation()
def CMD_CHANNEL(context: Context) -> "TYPE_MESSAGEABLE_CHANNEL":
    """This argument is the channel the command was sent in"""
    return context.channel


@define_annotation()
def CMD_ARGS(context: Context) -> list:
    """This argument is all of the arguments sent with this command"""
    return context.args



import logging
import re
from base64 import b64encode
from typing import Any, Dict, Union

import aiohttp  # type: ignore

from dis_snek.const import logger_name

log = logging.getLogger(logger_name)

try:
    import orjson as json
except ImportError:
    log.warning("orjson not installed, built-in json library will be used")
    import json


_quotes = {
    '"': '"',
    "‘": "’",
    "‚": "‛",
    "“": "”",
    "„": "‟",
    "⹂": "⹂",
    "「": "」",
    "『": "』",
    "〝": "〞",
    "﹁": "﹂",
    "﹃": "﹄",
    "＂": "＂",
    "｢": "｣",
    "«": "»",
    "‹": "›",
    "《": "》",
    "〈": "〉",
}
_pending_regex = r"(1.*2|[^\s]+)"
_pending_regex = _pending_regex.replace("1", f"[{''.join([k for k in _quotes.keys()])}]")
_pending_regex = _pending_regex.replace("2", f"[{''.join([k for k in _quotes.values()])}]")

arg_parse = re.compile(_pending_regex)
white_space = re.compile(r"\s+")
initial_word = re.compile(r"^([^\s]+)\s*?")


class OverriddenJson:
    @staticmethod
    def dumps(*args, **kwargs) -> str:
        data = json.dumps(*args, **kwargs)
        if isinstance(data, bytes):
            data = data.decode("utf-8")
        return data

    @staticmethod
    def loads(*args, **kwargs) -> dict:
        return json.loads(*args, **kwargs)


async def response_decode(response: aiohttp.ClientResponse) -> Union[Dict[str, Any], str]:
    """
    Return the response text in its correct format, be it dict, or string.

    Args:
        response: the aiohttp response
    Returns:
        the response text field in its correct type
    """
    text = await response.text(encoding="utf-8")

    if response.headers.get("content-type") == "application/json":
        return OverriddenJson.loads(text)
    return text


def get_args(text: str):
    """
    Get arguments from an input text.

    Args:
        text: The text to process
    Returns:
        A list of words
    """
    return arg_parse.findall(text)


def get_first_word(text: str):
    """
    Get a the first word in a string, regardless of whitespace type.

    Args:
        text: The text to process
    Returns:
         The requested word
    """
    found = initial_word.findall(text)
    if len(found) == 0:
        return None
    return found[0]


def _get_mime_type_for_image(data: bytes):
    # taken from d.py, alternative is to use libmagic, which would require users to install libs
    if data.startswith(b"\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"):
        return "image/png"
    elif data[0:3] == b"\xff\xd8\xff" or data[6:10] in (b"JFIF", b"Exif"):
        return "image/jpeg"
    elif data.startswith((b"\x47\x49\x46\x38\x37\x61", b"\x47\x49\x46\x38\x39\x61")):
        return "image/gif"
    elif data.startswith(b"RIFF") and data[8:12] == b"WEBP":
        return "image/webp"
    else:
        raise ValueError("Unsupported image type given")


def _bytes_to_base64_data(data: bytes) -> str:
    mime = _get_mime_type_for_image(data)
    return f"data:{mime};base64,{b64encode(data).decode('ascii')}"



import copy
import logging

from dis_snek.const import logger_name, MISSING
from dis_snek.event_processors._template import EventMixinTemplate
from dis_snek.models import listen, events
from dis_snek.models.events import RawGatewayEvent

log = logging.getLogger(logger_name)


class MemberEvents(EventMixinTemplate):
    @listen()
    async def _on_raw_guild_member_add(self, event: RawGatewayEvent) -> None:
        g_id = event.data.pop("guild_id")
        member = self.cache.place_member_data(g_id, event.data)
        self.dispatch(events.MemberAdd(g_id, member))

    @listen()
    async def _on_raw_guild_member_remove(self, event: RawGatewayEvent) -> None:
        g_id = event.data.pop("guild_id")
        user = self.cache.place_user_data(event.data["user"])
        self.dispatch(events.MemberAdd(g_id, await self.cache.get_member(g_id, user.id, False) or user))

    @listen()
    async def _on_raw_guild_member_update(self, event: RawGatewayEvent) -> None:
        g_id = event.data.pop("guild_id")
        before = copy.copy(await self.cache.get_member(g_id, event.data["user"]["id"], False)) or MISSING
        self.dispatch(events.MemberUpdate(g_id, before, self.cache.place_member_data(g_id, event.data)))



import asyncio
import datetime
import importlib.util
import inspect
import logging
import re
import sys
import traceback
from typing import TYPE_CHECKING, Callable, Coroutine, Dict, List, Optional, Union, Awaitable

import aiohttp

from dis_snek.const import logger_name, GLOBAL_SCOPE, MISSING, MENTION_PREFIX
from dis_snek.errors import (
    GatewayNotFound,
    SnakeException,
    WebSocketClosed,
    WebSocketRestart,
    BotException,
    ScaleLoadException,
    ExtensionLoadException,
    ExtensionNotFound,
    Forbidden,
    InteractionMissingAccess,
)
from dis_snek.event_processors import *
from dis_snek.gateway import WebsocketClient
from dis_snek.http_client import HTTPClient
from dis_snek.models import (
    Activity,
    Application,
    Guild,
    Listener,
    listen,
    Message,
    Scale,
    SnakeBotUser,
    User,
    Member,
    StickerPack,
    Sticker,
    events,
    InteractionCommand,
    SlashCommand,
    OptionTypes,
    MessageCommand,
    BaseCommand,
    to_snowflake,
    to_snowflake_list,
    ComponentContext,
    InteractionContext,
    MessageContext,
    AutocompleteContext,
    ComponentCommand,
    to_optional_snowflake,
    Context,
    application_commands_to_dict,
    sync_needed,
)
from dis_snek.models.discord_objects.components import get_components_ids, BaseComponent
from dis_snek.models.enums import ComponentTypes, Intents, InteractionTypes, Status, ActivityType
from dis_snek.models.events import RawGatewayEvent, MessageCreate
from dis_snek.models.events.internal import Component
from dis_snek.models.wait import Wait
from dis_snek.smart_cache import GlobalCache
from dis_snek.utils.input_utils import get_first_word, get_args
from dis_snek.utils.misc_utils import wrap_partial

if TYPE_CHECKING:
    from dis_snek.models import Snowflake_Type, TYPE_ALL_CHANNEL
    from asyncio import Future

log = logging.getLogger(logger_name)


class Snake(
    ChannelEvents,
    GuildEvents,
    MemberEvents,
    MessageEvents,
    ReactionEvents,
    RoleEvents,
    StageEvents,
    ThreadEvents,
    UserEvents,
):
    """
    The bot client.

    note:
        By default, all non-privileged intents will be enabled

    Attributes:
        intents Union[int, Intents]: The intents to use
        loop: An event loop to use, normally leave this blank
        default_prefix str: The default_prefix to use for message commands, defaults to your bot being mentioned
        get_prefix Callable[..., Coroutine]: A coroutine that returns a string to determine prefixes
        sync_interactions bool: Should application commands be synced with discord?
        delete_unused_application_cmds bool: Delete any commands from discord that arent implemented in this client
        asyncio_debug bool: Enable asyncio debug features
        status Status: The status the bot should login with (IE ONLINE, DND, IDLE)
        activity Activity: The activity the bot should login "playing"

    Optionally, you can configure the caches here, by specifying the name of the cache, followed by a dict-style object to use.
    It is recommended to use `smart_cache.create_cache` to configure the cache here.
    as an example, this is a recommended attribute `message_cache=create_cache(250, 50)`,

    !!! note
        Setting a message cache hard limit to None is not recommended, as it could result in extremely high memory usage, we suggest a sane limit.

    """

    def __init__(
        self,
        intents: Union[int, Intents] = Intents.DEFAULT,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        default_prefix: str = MENTION_PREFIX,
        get_prefix: Callable[..., Coroutine] = MISSING,
        sync_interactions: bool = False,
        delete_unused_application_cmds: bool = False,
        debug_scope: "Snowflake_Type" = MISSING,
        asyncio_debug: bool = False,
        status: Status = Status.ONLINE,
        activity: Union[Activity, str] = None,
        **kwargs,
    ):

        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop() if loop is None else loop

        # Configuration

        if asyncio_debug:
            log.warning("Asyncio Debug is enabled, Your log will contain additional errors and warnings")
            import tracemalloc

            tracemalloc.start()
            self.loop.set_debug(True)

        self.intents = intents
        """The intents in use"""
        self.sync_interactions = sync_interactions
        """Should application commands be synced"""
        self.del_unused_app_cmd: bool = delete_unused_application_cmds
        """Should unused application commands be deleted?"""
        self.debug_scope = to_optional_snowflake(debug_scope)
        """Sync global commands as guild for quicker command updates during debug"""
        self.default_prefix = default_prefix
        """The default prefix to be used for message commands"""
        self.get_prefix = get_prefix if get_prefix is not MISSING else self.get_prefix
        """A coroutine that returns a prefix, for dynamic prefixes"""

        # resources

        self.http: HTTPClient = HTTPClient(loop=self.loop)
        """The HTTP client to use when interacting with discord endpoints"""
        self.ws: WebsocketClient = MISSING
        """The websocket collection for the Discord Gateway."""

        # flags
        self._ready = False
        self._closed = False
        self._guild_event = asyncio.Event()
        self.guild_event_timeout = 3
        """How long to wait for guilds to be cached"""
        self.start_time = MISSING
        """The DateTime the bot started at"""

        self._mention_reg = MISSING

        # caches
        self.cache: GlobalCache = GlobalCache(self, **{k: v for k, v in kwargs.items() if hasattr(GlobalCache, k)})
        # these store the last sent presence data for change_presence
        self._status: Status = status
        if isinstance(activity, str):
            self._activity = Activity.create(name=str(activity))
        else:
            self._activity: Activity = activity

        self._user: SnakeBotUser = MISSING
        self._app: Application = MISSING

        # collections

        self.commands: Dict[str, MessageCommand] = {}
        """A dictionary of registered commands: `{name: command}`"""
        self.interactions: Dict["Snowflake_Type", Dict[str, InteractionCommand]] = {}
        """A dictionary of registered application commands: `{cmd_id: command}`"""
        self._component_callbacks: Dict[str, Callable[..., Coroutine]] = {}
        self._interaction_scopes: Dict["Snowflake_Type", "Snowflake_Type"] = {}
        self.__extensions = {}
        self.scales = {}
        """A dictionary of mounted Scales"""
        self.listeners: Dict[str, List] = {}
        self.waits: Dict[str, List] = {}

    @property
    def is_closed(self) -> bool:
        """Is the bot closed?"""
        return self._closed

    @property
    def is_ready(self):
        return self._ready

    @property
    def latency(self) -> float:
        """Returns the latency of the websocket connection"""
        return self.ws.latency

    @property
    def user(self) -> SnakeBotUser:
        """Returns the bot's user"""
        return self._user

    @property
    def app(self) -> Application:
        """Returns the bots application"""
        return self._app

    @property
    def owner(self) -> Optional["User"]:
        """Returns the bot's owner'"""
        try:
            return self.app.owner
        except TypeError:
            return MISSING

    @property
    def guilds(self) -> List["Guild"]:
        return self.user.guilds

    @property
    def status(self) -> Status:
        """Get the status of the bot. IE online, afk, dnd"""
        return self._status

    @property
    def activity(self) -> Activity:
        """Get the activity of the bot"""
        return self._activity

    @property
    def application_commands(self):
        """a list of all application commands registered within the bot"""
        commands = []
        for scope in self.interactions.keys():
            for cmd in self.interactions[scope].values():
                if cmd not in commands:
                    commands.append(cmd)

        return commands

    async def get_prefix(self, message: Message) -> str:
        """A method to get the bot's default_prefix, can be overridden to add dynamic prefixes.

        !!! note
            To easily override this method, simply use the `get_prefix` parameter when instantiating the client

        Args:
            message: A message to determine the prefix from.

        Returns:
            A string to use as a prefix, by default will return `client.default_prefix`
        """
        return self.default_prefix

    async def login(self, token):
        """
        Login to discord

        Args:
            token str: Your bot's token
        """
        # i needed somewhere to put this call,
        # login will always run after initialisation
        # so im gathering commands here
        self._gather_commands()

        log.debug(f"Attempting to login")
        me = await self.http.login(token.strip())
        self._user = SnakeBotUser.from_dict(me, self)
        self.cache.place_user_data(me)
        self._app = Application.from_dict(await self.http.get_current_bot_information(), self)
        self._mention_reg = re.compile(fr"^(<@!?{self.user.id}*>\s)")
        self.start_time = datetime.datetime.now()
        self.dispatch(events.Login())
        await self._ws_connect()

    async def _ws_connect(self):
        params = {
            "http": self.http,
            "dispatch": self.dispatch,
            "intents": self.intents,
            "resume": False,
            "session_id": None,
            "sequence": None,
            "presence": {"status": self._status, "activities": [self._activity.to_dict()] if self._activity else []},
        }
        while not self.is_closed:
            log.info(f"Attempting to {'re' if params['resume'] else ''}connect to gateway...")

            try:
                self.ws = await WebsocketClient.connect(**params)

                await self.ws.run()
            except WebSocketRestart as ex:
                # internally requested restart
                self.dispatch(events.Disconnect())
                if ex.resume:
                    params.update(resume=True, session_id=self.ws.session_id, sequence=self.ws.sequence)
                    continue
                params.update(resume=False, session_id=None, sequence=None)

            except (OSError, GatewayNotFound, aiohttp.ClientError, asyncio.TimeoutError, WebSocketClosed) as ex:
                log.debug("".join(traceback.format_exception(type(ex), ex, ex.__traceback__)))
                self.dispatch(events.Disconnect())

                if isinstance(ex, WebSocketClosed):
                    if ex.code == 1000:
                        if self._ready:
                            # the bot disconnected, attempt to reconnect to gateway
                            params.update(resume=True, session_id=self.ws.session_id, sequence=self.ws.sequence)
                            continue
                        else:
                            return
                    elif ex.code == 4011:
                        raise SnakeException("Your bot is too large, you must use shards") from None
                    elif ex.code == 4013:
                        raise SnakeException("Invalid Intents have been passed") from None
                    elif ex.code == 4014:
                        raise SnakeException(
                            "You have requested privileged intents that have not been enabled or approved. Check the developer dashboard"
                        ) from None
                    raise

                if isinstance(ex, OSError) and ex.errno in (54, 10054):
                    print("should reconnect")
                    params.update(resume=True, session_id=self.ws.session_id, sequence=self.ws.sequence)
                    continue
                params.update(resume=False, session_id=None, sequence=None)

            except Exception as e:
                self.dispatch(events.Disconnect())
                log.error("".join(traceback.format_exception(type(e), e, e.__traceback__)))
                params.update(resume=False, session_id=None, sequence=None)

            await asyncio.sleep(5)

    def _queue_task(self, coro, event, *args, **kwargs):
        async def _async_wrap(_coro, _event, *_args, **_kwargs):
            try:
                if len(_event.__attrs_attrs__) == 1:
                    await _coro()
                else:
                    await _coro(_event, *_args, **_kwargs)
            except asyncio.CancelledError:
                pass
            except Exception as e:
                await self.on_error(event, e)

        wrapped = _async_wrap(coro, event, *args, **kwargs)

        return asyncio.create_task(wrapped, name=f"snake:: {event.resolved_name}")

    async def on_error(self, source: str, error: Exception, *args, **kwargs) -> None:
        """
        Catches all errors dispatched by the library.

        By default it will format and print them to console

        Override this to change error handling behaviour
        """
        print(f"Ignoring exception in {source}:\n{traceback.format_exc()}", file=sys.stderr)

    async def on_command_error(self, source: str, error: Exception, *args, **kwargs) -> None:
        """
        Catches all errors dispatched by commands

        By default it will call `Snake.on_error`

        Override this to change error handling behavior
        """
        return await self.on_error(source, error, *args, **kwargs)

    async def on_command(self, ctx: Context) -> None:
        """
        Called *after* any command is ran

        By default, it will simply log the command, override this to change that behaviour
        Args:
            ctx: The context of the command that was called
        """
        if isinstance(ctx, MessageContext):
            symbol = "@"
        elif isinstance(ctx, ComponentContext):
            symbol = "¢"
        elif isinstance(ctx, InteractionContext):
            symbol = "/"
        else:
            symbol = "?"  # likely custom context
        log.info(f"Command Called: {symbol}{ctx.invoked_name} with {ctx.args = } | {ctx.kwargs = }")

    @listen()
    async def _on_websocket_ready(self, event: events.RawGatewayEvent) -> None:
        """
        Catches websocket ready and determines when to dispatch the client `READY` signal.

        Args:
            event: The websocket ready packet
        """
        data = event.data
        expected_guilds = set(to_snowflake(guild["id"]) for guild in data["guilds"])
        self._user._add_guilds(expected_guilds)

        while True:
            try:  # wait to let guilds cache
                await asyncio.wait_for(self._guild_event.wait(), self.guild_event_timeout)
            except asyncio.TimeoutError:
                log.warning("Timeout waiting for guilds cache: Not all guilds will be in cache")
                break
            self._guild_event.clear()

            if len(self.cache.guild_cache) == len(expected_guilds):
                # all guilds cached
                break

        # cache slash commands
        await self._init_interactions()

        self._ready = True
        self.dispatch(events.Ready())

    def start(self, token):
        """
        Start the bot.

        info:
            This is the recommended method to start the bot

        Args:
            token str: Your bot's token
        """
        try:
            self.loop.run_until_complete(self.login(token))
        except KeyboardInterrupt:
            self.loop.run_until_complete(self.stop())

    async def stop(self):
        log.debug("Stopping the bot.")
        self._ready = False
        await self.ws.close(1001)

    def dispatch(self, event: events.BaseEvent, *args, **kwargs):
        """
        Dispatch an event.

        Args:
            event: The event to be dispatched.
        """
        log.debug(f"Dispatching Event: {event.resolved_name}")
        listeners = self.listeners.get(event.resolved_name, [])
        for _listen in listeners:
            try:
                self._queue_task(_listen, event, *args, **kwargs)
            except Exception as e:
                raise BotException(f"An error occurred attempting during {event.resolved_name} event processing")

        _waits = self.waits.get(event.resolved_name, [])
        index_to_remove = []
        for i, _wait in enumerate(_waits):
            result = _wait(event)
            if result:
                index_to_remove.append(i)

        for idx in index_to_remove:
            _waits.pop(idx)

    def wait_for(self, event: str, checks: Optional[Callable[..., bool]] = MISSING, timeout: Optional[float] = None):
        """
        Waits for a WebSocket event to be dispatched.

        Args:
            event: The name of event to wait.
            checks: A predicate to check what to wait for.
            timeout: The number of seconds to wait before timing out.

        Returns:
            The event object.
        """
        if event not in self.waits:
            self.waits[event] = []

        future = self.loop.create_future()
        self.waits[event].append(Wait(event, checks, future))

        return asyncio.wait_for(future, timeout)

    async def wait_for_component(
        self,
        messages: Union[Message, int, list] = None,
        components: Optional[
            Union[List[List[Union["BaseComponent", dict]]], List[Union["BaseComponent", dict]], "BaseComponent", dict]
        ] = None,
        check=None,
        timeout=None,
    ) -> Awaitable["Future"]:
        """
        Waits for a message to be sent to the bot.

        Args:
            messages: The message object to check for.
            components: The components to wait for.
            check: A predicate to check what to wait for.
            timeout: The number of seconds to wait before timing out.

        Returns:
            `Component` that was invoked, or `None` if timed out. Use `.context` to get the `ComponentContext`.
        """
        if not (messages or components):
            raise ValueError("You must specify messages or components (or both)")

        message_ids = (
            to_snowflake_list(messages) if isinstance(messages, list) else to_snowflake(messages) if messages else None
        )
        custom_ids = list(get_components_ids(components)) if components else None

        # automatically convert improper custom_ids
        if custom_ids and not all(isinstance(x, str) for x in custom_ids):
            custom_ids = [str(i) for i in custom_ids]

        def _check(event: Component):
            ctx: ComponentContext = event.context
            # if custom_ids is empty or there is a match
            wanted_message = not message_ids or ctx.message.id in (
                [message_ids] if isinstance(message_ids, int) else message_ids
            )
            wanted_component = not custom_ids or ctx.custom_id in custom_ids
            if wanted_message and wanted_component:
                if check is None or check(event):
                    return True
                return False
            return False

        return await self.wait_for("component", checks=_check, timeout=timeout)

    def add_listener(self, listener: Listener):
        """
        Add a listener for an event, if no event is passed, one is determined

        Args:
            coro Listener: The listener to add to the client
        """
        if listener.event not in self.listeners:
            self.listeners[listener.event] = []
        self.listeners[listener.event].append(listener)

    def add_interaction(self, command: InteractionCommand):
        """
        Add a slash command to the client.

        Args:
            command InteractionCommand: The command to add
        """
        if self.debug_scope:
            command.scopes = [self.debug_scope]
        for scope in command.scopes:

            if scope not in self.interactions:
                self.interactions[scope] = {}
            elif command.resolved_name in self.interactions[scope]:
                old_cmd = self.interactions[scope][command.resolved_name]
                raise ValueError(f"Duplicate Command! {scope}::{old_cmd.resolved_name}")

            self.interactions[scope][command.resolved_name] = command

    def add_message_command(self, command: MessageCommand):
        """
        Add a message command to the client.

        Args:
            command InteractionCommand: The command to add
        """
        if command.name not in self.commands:
            self.commands[command.name] = command
            return
        raise ValueError(f"Duplicate Command! Multiple commands share the name `{command.name}`")

    def add_component_callback(self, command: ComponentCommand):
        """Add a component callback to the client

        Args:
            command: The command to add
        """
        for listener in command.listeners:
            # I know this isn't an ideal solution, but it means we can lookup callbacks with O(1)
            if listener not in self._component_callbacks.keys():
                self._component_callbacks[listener] = command
                continue
            else:
                raise ValueError(f"Duplicate Component! Multiple component callbacks for `{listener}`")

    def _gather_commands(self):
        """Gathers commands from __main__ and self"""

        def process(_cmds):

            for func in _cmds:
                if isinstance(func, ComponentCommand):
                    self.add_component_callback(func)
                elif isinstance(func, InteractionCommand):
                    self.add_interaction(func)
                elif isinstance(func, MessageCommand):
                    self.add_message_command(func)
                elif isinstance(func, Listener):
                    self.add_listener(func)

            log.debug(f"{len(_cmds)} commands have been loaded from `__main__` and `client`")

        process(
            [obj for _, obj in inspect.getmembers(sys.modules["__main__"]) if isinstance(obj, (BaseCommand, Listener))]
        )
        process(
            [wrap_partial(obj, self) for _, obj in inspect.getmembers(self) if isinstance(obj, (BaseCommand, Listener))]
        )

    async def _init_interactions(self) -> None:
        """
        Initialise slash commands.

        If `sync_interactions` this will submit all registered slash commands to discord.
        Otherwise, it will get the list of interactions and cache their scopes.
        """
        # allow for cogs and main to share the same decorator
        try:
            if self.sync_interactions:
                await self.synchronise_interactions()
            else:
                await self._cache_interactions(warn_missing=True)
        except Exception as e:
            await self.on_error("Interaction Syncing", e)

    async def _cache_interactions(self, warn_missing: bool = False):
        """Get all interactions used by this bot and cache them."""
        bot_scopes = set(g.id for g in self.cache.guild_cache.values())
        bot_scopes.add(GLOBAL_SCOPE)

        # Match all interaction is registered with discord's data.
        for scope in self.interactions:
            bot_scopes.discard(scope)
            try:
                remote_cmds = await self.http.get_interaction_element(self.user.id, scope)
            except Forbidden as e:
                raise InteractionMissingAccess(scope) from None

            remote_cmds = {cmd_data["name"]: cmd_data for cmd_data in remote_cmds}
            found = set()  # this is a temporary hack to fix subcommand detection
            for cmd in self.interactions[scope].values():
                cmd_data = remote_cmds.get(cmd.name, MISSING)
                if cmd_data is MISSING:
                    if cmd.name not in found:
                        if warn_missing:
                            log.error(
                                f'Detected yet to sync slash command "/{cmd.name}" for scope '
                                f"{'global' if scope == GLOBAL_SCOPE else scope}"
                            )
                    continue
                else:
                    found.add(cmd.name)

                self._interaction_scopes[str(cmd_data["id"])] = scope
                cmd.cmd_id = str(cmd_data["id"])

            if warn_missing:
                for cmd_data in remote_cmds.values():
                    log.error(
                        f"Detected unimplemented slash command \"/{cmd_data['name']}\" for scope "
                        f"{'global' if scope == GLOBAL_SCOPE else scope}"
                    )

        # Remaining guilds that bot is in but, no interaction is registered
        for scope in bot_scopes:
            try:
                remote_cmds = await self.http.get_interaction_element(self.user.id, scope)
            except Forbidden:
                # We will just assume they don't want application commands in this guild.
                log.debug(f"Bot was not invited to guild {scope} with `application.commands` scope")
                continue

            for cmd_data in remote_cmds:
                self._interaction_scopes[str(cmd_data["id"])] = scope
                if warn_missing:
                    log.error(
                        f"Detected unimplemented slash command \"/{cmd_data['name']}\" for scope "
                        f"{'global' if scope == GLOBAL_SCOPE else scope}"
                    )

    async def synchronise_interactions(self) -> None:
        """Synchronise registered interactions with discord

        One flaw of this is it cant determine if context menus need updating,
        as discord isn't returning that data on get req, so they are unnecessarily updated"""

        # first we need to make sure our local copy of cmd_ids is up-to-date
        await self._cache_interactions()
        cmd_scopes = [to_snowflake(g_id) for g_id in self._user._guild_ids] + [GLOBAL_SCOPE]

        guild_perms = {}

        cmds_json = application_commands_to_dict(self.interactions)

        for cmd_scope in cmd_scopes:
            try:
                cmds_resp_data = await self.http.get_interaction_element(self.user.id, cmd_scope)
                need_to_sync = False
                cmds_to_sync = []
                found = []

                for local_cmd in self.interactions.get(cmd_scope, {}).values():
                    # try and find remote equiv of this command
                    remote_cmd = next((v for v in cmds_resp_data if v["id"] == local_cmd.cmd_id), None)

                    local_cmd = next((c for c in cmds_json[cmd_scope] if c["name"] == local_cmd.name))

                    if local_cmd not in cmds_to_sync:
                        cmds_to_sync.append(local_cmd)
                    if remote_cmd not in found:
                        found.append(remote_cmd)

                    # todo: prevent un-needed syncs for subcommands
                    if sync_needed(local_cmd, remote_cmd):
                        # if command local data doesnt match remote, a change has been made, sync it
                        need_to_sync = True

                if need_to_sync:
                    log.info(f"Updating {len(cmds_to_sync)} commands in {cmd_scope}")
                    cmd_sync_resp = await self.http.post_interaction_element(
                        self.user.id, cmds_to_sync, guild_id=cmd_scope
                    )
                    # cache cmd_ids and their scopes
                    for cmd_data in cmd_sync_resp:
                        self._interaction_scopes[cmd_data["id"]] = cmd_scope
                        if cmd_data["name"] in self.interactions[cmd_scope]:
                            self.interactions[cmd_scope][cmd_data["name"]].cmd_id = str(cmd_data["id"])
                        else:
                            # sub_cmd
                            for sc in cmd_data["options"]:
                                if sc["type"] == OptionTypes.SUB_COMMAND:
                                    if f"{cmd_data['name']} {sc['name']}" in self.interactions[cmd_scope]:
                                        self.interactions[cmd_scope][f"{cmd_data['name']} {sc['name']}"].cmd_id = str(
                                            cmd_data["id"]
                                        )
                                elif sc["type"] == OptionTypes.SUB_COMMAND_GROUP:
                                    for _sc in sc["options"]:
                                        if (
                                            f"{cmd_data['name']} {sc['name']} {_sc['name']}"
                                            in self.interactions[cmd_scope]
                                        ):
                                            self.interactions[cmd_scope][
                                                f"{cmd_data['name']} {sc['name']} {_sc['name']}"
                                            ].cmd_id = str(cmd_data["id"])

                else:
                    log.debug(f"{cmd_scope} is already up-to-date with {len(cmds_resp_data)} commands.")

                for local_cmd in self.interactions.get(cmd_scope, {}).values():
                    if not local_cmd.permissions:
                        continue
                    for perm in local_cmd.permissions:
                        if perm.guild_id not in guild_perms:
                            guild_perms[perm.guild_id] = []
                        guild_perms[perm.guild_id].append(
                            {"id": local_cmd.cmd_id, "permissions": [perm.to_dict() for perm in local_cmd.permissions]}
                        )

                for perm_scope in guild_perms:
                    log.debug(f"Updating {len(guild_perms[perm_scope])} command permissions in {perm_scope}")
                    await self.http.batch_edit_application_command_permissions(
                        application_id=self.user.id, scope=perm_scope, data=guild_perms[perm_scope]
                    )

                if self.del_unused_app_cmd:
                    for cmd in [c for c in cmds_resp_data if c not in found]:
                        scope = cmd.get("guild_id", GLOBAL_SCOPE)
                        log.warning(
                            f"Deleting unimplemented slash command \"/{cmd['name']}\" from scope "
                            f"{'global' if scope == GLOBAL_SCOPE else scope}"
                        )
                        await self.http.delete_interaction_element(
                            self.user.id, cmd.get("guild_id", GLOBAL_SCOPE), cmd["id"]
                        )
            except Forbidden as e:
                raise InteractionMissingAccess(cmd_scope) from None

    async def get_context(
        self, data: Union[dict, Message], interaction: bool = False
    ) -> Union[MessageContext, InteractionContext, ComponentContext, AutocompleteContext]:
        """
        Return a context object based on data passed

        note:
            If you want to use custom context objects, this is the method to override. Your replacement must take the same arguments as this, and return a Context-like object.

        Args:
            data: The data of the event
            interaction: Is this an interaction or not?

        returns:
            Context object
        """
        # this line shuts up IDE warnings
        cls: Union[MessageContext, ComponentContext, InteractionContext, AutocompleteContext]

        if interaction:
            # todo: change to match
            if data["type"] == InteractionTypes.MESSAGE_COMPONENT:
                return ComponentContext.from_dict(data, self)
            elif data["type"] == InteractionTypes.AUTOCOMPLETE:
                cls = AutocompleteContext.from_dict(data, self)
            else:
                cls = InteractionContext.from_dict(data, self)

            invoked_name: str = data["data"]["name"]
            kwargs = {}

            if options := data["data"].get("options"):
                o_type = options[0]["type"]
                if o_type in (OptionTypes.SUB_COMMAND, OptionTypes.SUB_COMMAND_GROUP):
                    # this is a subcommand, process accordingly
                    if o_type == OptionTypes.SUB_COMMAND:
                        invoked_name = f"{invoked_name} {options[0]['name']}"
                        options = options[0].get("options", [])
                    else:
                        invoked_name = (
                            f"{invoked_name} {options[0]['name']} "
                            f"{next(x for x in options[0]['options'] if x['type'] == OptionTypes.SUB_COMMAND)['name']}"
                        )
                        options = options[0]["options"][0].get("options", [])

                for option in options:
                    value = option.get("value")

                    # todo change to match statement
                    # this block here resolves the options using the cache
                    if option["type"] == OptionTypes.USER:
                        value = (
                            self.cache.member_cache.get((to_snowflake(data.get("guild_id", 0)), to_snowflake(value)))
                            or self.cache.user_cache.get(to_snowflake(value))
                        ) or value
                    elif option["type"] == OptionTypes.CHANNEL:
                        value = self.cache.channel_cache.get(to_snowflake(value)) or value
                    elif option["type"] == OptionTypes.ROLE:
                        value = self.cache.role_cache.get(to_snowflake(value)) or value
                    elif option["type"] == OptionTypes.MENTIONABLE:
                        snow = to_snowflake(value)
                        if user := self.cache.member_cache.get(snow) or self.cache.user_cache.get(snow):
                            value = user
                        elif role := self.cache.role_cache.get(snow):
                            value = role
                    if option.get("focused", False):
                        cls.focussed_option = option.get("name")
                    kwargs[option["name"].lower()] = value

            cls.invoked_name = invoked_name
            cls.kwargs = kwargs
            cls.args = [v for v in kwargs.values()]

            return cls
        else:
            return MessageContext.from_message(self, data)

    @listen("raw_interaction_create")
    async def _dispatch_interaction(self, event: RawGatewayEvent) -> None:
        """
        Identify and dispatch interaction of slash commands or components.

        Args:
            raw interaction event
        """
        interaction_data = event.data

        if interaction_data["type"] in (
            InteractionTypes.PING,
            InteractionTypes.APPLICATION_COMMAND,
            InteractionTypes.AUTOCOMPLETE,
        ):
            interaction_id = interaction_data["data"]["id"]
            name = interaction_data["data"]["name"]
            scope = self._interaction_scopes.get(str(interaction_id))

            if scope in self.interactions:
                ctx = await self.get_context(interaction_data, True)

                command: SlashCommand = self.interactions[scope][ctx.invoked_name]  # type: ignore
                log.debug(f"{scope} :: {command.name} should be called")

                if auto_opt := getattr(ctx, "focussed_option", None):
                    await command.autocomplete_callbacks[auto_opt](ctx, **ctx.kwargs)
                else:
                    try:
                        await command(ctx, **ctx.kwargs)
                    except Exception as e:
                        await self.on_command_error(f"cmd /`{name}`", e)
                    finally:
                        await self.on_command(ctx)
            else:
                log.error(f"Unknown cmd_id received:: {interaction_id} ({name})")

        elif interaction_data["type"] == InteractionTypes.MESSAGE_COMPONENT:
            # Buttons, Selects, ContextMenu::Message
            ctx = await self.get_context(interaction_data, True)
            component_type = interaction_data["data"]["component_type"]

            self.dispatch(events.Component(ctx))
            if callback := self._component_callbacks.get(ctx.custom_id):
                try:
                    await callback(ctx)
                except Exception as e:
                    await self.on_command_error(f"Component Callback for {ctx.custom_id}", e)
                finally:
                    await self.on_command(ctx)
            if component_type == ComponentTypes.BUTTON:
                self.dispatch(events.Button(ctx))
            if component_type == ComponentTypes.SELECT:
                self.dispatch(events.Select(ctx))

        else:
            raise NotImplementedError(f"Unknown Interaction Received: {interaction_data['type']}")

    @listen("message_create")
    async def _dispatch_msg_commands(self, event: MessageCreate):
        """Determine if a command is being triggered, and dispatch it."""
        message = event.message

        if not message.author.bot:
            prefix = await self.get_prefix(message)

            if prefix == MENTION_PREFIX:
                mention = self._mention_reg.search(message.content)
                if mention:
                    prefix = mention.group()
                else:
                    return

            if message.content.startswith(prefix):
                invoked_name = get_first_word(message.content.removeprefix(prefix))
                command = self.commands.get(invoked_name)
                if command and command.enabled:
                    context = await self.get_context(message)
                    context.invoked_name = invoked_name
                    context.prefix = prefix
                    context.args = get_args(context.content_parameters)
                    try:
                        await command(context)
                    except Exception as e:
                        await self.on_command_error(f"cmd `{invoked_name}`", e)
                    finally:
                        await self.on_command(context)

    def get_scale(self, name) -> Optional[Scale]:
        """
        Get a scale
        Args:
            name: The name of the scale, or the name of it's extension

        Returns:
            Scale or None if no scale is found
        """
        if name not in self.scales.keys():
            for scale in self.scales.values():
                if scale.extension_name == name:
                    return scale

        return self.scales.get(name, None)

    def grow_scale(self, file_name: str, package: str = None) -> None:
        """
        A helper method to load a scale

        Args:
            file_name: The name of the file to load the scale from.
            package: The package this scale is in.
        """
        self.load_extension(file_name, package)

    def shed_scale(self, scale_name: str) -> None:
        """
        Helper method to unload a scale

        Args:
            scale_name: The name of the scale to unload.
        """
        if scale := self.get_scale(scale_name):
            return self.unload_extension(inspect.getmodule(scale).__name__)

        raise ScaleLoadException(f"Unable to shed scale: No scale exists with name: `{scale_name}`")

    def regrow_scale(self, scale_name: str) -> None:
        """
        Helper method to reload a scale.

        Args:
            scale_name: The name of the scale to reload
        """
        self.shed_scale(scale_name)
        self.grow_scale(scale_name)

    def load_extension(self, name: str, package: str = None):
        """
        Load an extension.

        Args:
            name: The name of the extension.
            package: The package the extension is in
        """
        name = importlib.util.resolve_name(name, package)
        if name in self.__extensions:
            raise Exception(f"{name} already loaded")

        module = importlib.import_module(name, package)
        try:
            setup = getattr(module, "setup")
            setup(self)
        except Exception as e:
            del sys.modules[name]
            raise ExtensionLoadException(f"Error loading {name}") from e

        else:
            log.debug(f"Loaded Extension: {name}")
            self.__extensions[name] = module
            return

    def unload_extension(self, name, package=None):
        """
        unload an extension.

        Args:
            name: The name of the extension.
            package: The package the extension is in
        """
        name = importlib.util.resolve_name(name, package)
        module = self.__extensions.get(name)

        if module is None:
            raise ExtensionNotFound(f"No extension called {name} is loaded")

        try:
            teardown = getattr(module, "teardown")
            teardown()
        except AttributeError:
            pass

        if scale := self.get_scale(name):
            scale.shed()

        del sys.modules[name]
        del self.__extensions[name]

    def reload_extension(self, name, package=None):
        """
        Helper method to reload an extension.
        Simply unloads, then loads the extension.

        Args:
            name: The name of the extension.
            package: The package the extension is in
        """
        name = importlib.util.resolve_name(name, package)
        module = self.__extensions.get(name)

        if module is None:
            log.warning("Attempted to reload extension thats not loaded. Loading extension instead")
            return self.load_extension(name, package)

        self.unload_extension(name, package)
        self.load_extension(name, package)

        # todo: maybe add an ability to revert to the previous version if unable to load the new one

    async def get_guild(self, guild_id: "Snowflake_Type") -> Guild:
        """
        Get a guild

        Note:
            This method is an alias for the cache which will either return a cached object, or query discord for the object
            if its not already cached.

        Args:
            guild_id: The ID of the guild to get

        Returns:
            Guild Object
        """
        return await self.cache.get_guild(guild_id)

    async def get_channel(self, channel_id: "Snowflake_Type") -> "TYPE_ALL_CHANNEL":
        """
        Get a channel

        Note:
            This method is an alias for the cache which will either return a cached object, or query discord for the object
            if its not already cached.

        Args:
            channel_id: The ID of the channel to get

        Returns:
            Channel Object
        """
        return await self.cache.get_channel(channel_id)

    async def get_user(self, user_id: "Snowflake_Type") -> User:
        """
        Get a user

        Note:
            This method is an alias for the cache which will either return a cached object, or query discord for the object
            if its not already cached.

        Args:
            user_id: The ID of the user to get

        Returns:
            User Object
        """
        return await self.cache.get_user(user_id)

    async def get_member(self, user_id: "Snowflake_Type", guild_id: "Snowflake_Type") -> Member:
        """
        Get a member from a guild

        Note:
            This method is an alias for the cache which will either return a cached object, or query discord for the object
            if its not already cached.

        Args:
            user_id: The ID of the member
            guild_id: The ID of the guild to get the member from

        Returns:
            Member object
        """
        return await self.cache.get_member(guild_id, user_id)

    async def get_sticker(self, sticker_id: "Snowflake_Type"):
        sticker_data = await self.http.get_sticker(sticker_id)
        return Sticker.from_dict(sticker_data, self)

    async def get_nitro_packs(self) -> List["StickerPack"]:
        packs_data = await self.http.list_nitro_sticker_packs()
        packs = []
        for pack_data in packs_data:
            packs.append(StickerPack.from_dict(pack_data, self))
        return packs

    async def change_presence(
        self, status: Optional[Union[str, Status]] = Status.ONLINE, activity: Optional[Union[Activity, str]] = None
    ):
        """
        Change the bots presence.

        Args:
            status: The status for the bot to be. i.e. online, afk, etc.
            activity: The activity for the bot to be displayed as doing.

        note::
            Bots may only be `playing` `streaming` or `listening`, other activity types are likely to fail.
        """
        if activity:
            if not isinstance(activity, Activity):
                # squash whatever the user passed into an activity
                activity = Activity.create(name=str(activity))

            if activity.type == ActivityType.STREAMING:
                if not activity.url:
                    log.warning("Streaming activity cannot be set without a valid URL attribute")
            elif activity.type not in [ActivityType.GAME, ActivityType.STREAMING, ActivityType.LISTENING]:
                log.warning(f"Activity type `{ActivityType(activity.type).name}` may not be enabled for bots")
        else:
            activity = self._activity if self._activity else []

        if status:
            if not isinstance(status, Status):
                try:
                    status = Status[status.upper()]
                except KeyError:
                    raise ValueError(f"`{status}` is not a valid status type. Please use the Status enum") from None
        else:
            # in case the user set status to None
            if self._status:
                status = self._status
            else:
                log.warning("Status must be set to a valid status type, defaulting to online")
                status = Status.ONLINE

        self._status = status
        self._activity = activity
        await self.ws.change_presence(activity.to_dict() if activity else None, status)



import logging
from typing import Union

from dis_snek.const import logger_name
from dis_snek.event_processors._template import EventMixinTemplate
from dis_snek.models import listen, events, User, Member, BaseChannel, Timestamp, to_snowflake, Activity
from dis_snek.models.enums import Status
from dis_snek.models.events import RawGatewayEvent

log = logging.getLogger(logger_name)


class UserEvents(EventMixinTemplate):
    @listen()
    async def _on_raw_typing_start(self, event: RawGatewayEvent) -> None:
        """
        Process raw typing start and dispatch a processed typing event.

        Args:
            event: raw typing start event
        """
        author: Union[User, Member]
        channel: BaseChannel
        guild = None

        if member := event.data.get("member"):
            author = self.cache.place_member_data(event.data.get("guild_id"), member)
            guild = await self.cache.get_guild(event.data.get("guild_id"))
        else:
            author = await self.cache.get_user(event.data.get("user_id"))

        channel = await self.cache.get_channel(event.data.get("channel_id"))

        self.dispatch(
            events.TypingStart(
                author=author,
                channel=channel,
                guild=guild,
                timestamp=Timestamp.utcfromtimestamp(event.data.get("timestamp")),
            )
        )

    @listen()
    async def _on_raw_presence_update(self, event: RawGatewayEvent) -> None:
        g_id = to_snowflake(event.data["guild_id"])
        user = await self.cache.get_user(event.data["user"]["id"])
        status = Status[event.data["status"].upper()]
        activities = [Activity.from_dict(a) for a in event.data.get("activities")]

        self.dispatch(events.PresenceUpdate(user, status, activities, event.data.get("client_status", None), g_id))



from .channel_events import ChannelEvents
from .guild_events import GuildEvents
from .member_events import MemberEvents
from .message_events import MessageEvents
from .reaction_events import ReactionEvents
from .role_events import RoleEvents
from .stage_events import StageEvents
from .thread_events import ThreadEvents
from .user_events import UserEvents



from typing import TYPE_CHECKING, List, Optional, Dict, Any, Union

import attr

from dis_snek.models.discord_objects.asset import Asset
from dis_snek.models.discord import DiscordObject, ClientObject
from dis_snek.models.enums import TeamMembershipState
from dis_snek.models.snowflake import to_snowflake
from dis_snek.utils.attr_utils import define

if TYPE_CHECKING:
    from dis_snek.models.discord_objects.user import User
    from dis_snek.models.snowflake import Snowflake_Type, SnowflakeObject
    from dis_snek.client import Snake


@define()
class TeamMember(ClientObject):
    membership_state: TeamMembershipState = attr.ib(converter=TeamMembershipState)
    # permissions: List[str] = attr.ib(default=["*"])  # disabled until discord adds more team roles
    team_id: "Snowflake_Type" = attr.ib()
    user: "User" = attr.ib()  # TODO: cache partial user (avatar, discrim, id, username)

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        data["user"] = client.cache.place_user_data(data["user"])
        return data


@define()
class Team(DiscordObject):
    icon: Optional[Asset] = attr.ib(default=None)
    members: List[TeamMember] = attr.ib(factory=list)
    name: str = attr.ib()
    owner_user_id: "Snowflake_Type" = attr.ib(converter=to_snowflake)

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        data["members"] = [TeamMember.from_dict(member, client) for member in data["members"]]
        if data["icon"]:
            data["icon"] = Asset.from_path_hash(client, f"team-icons/{data['id']}/{{}}.png", data["icon"])
        return data

    @property
    def owner(self) -> "User":
        return self._client.cache.user_cache.get(self.owner_user_id)

    def is_in_team(self, user: Union["SnowflakeObject", "Snowflake_Type"]) -> bool:
        """
        Returns True if the passed user or ID is a member within the team.
        Args:
            user: The user or user ID to check

        Returns:
            Boolean indicating whether the user is in the team
        """
        return to_snowflake(user) in [m.id for m in self.members]



import copy
import logging

from dis_snek.const import logger_name
from dis_snek.event_processors._template import EventMixinTemplate
from dis_snek.models import listen, events, to_snowflake
from dis_snek.models.events import RawGatewayEvent

log = logging.getLogger(logger_name)


class MessageEvents(EventMixinTemplate):
    @listen()
    async def _on_raw_message_create(self, event: RawGatewayEvent) -> None:
        """
        Automatically convert MESSAGE_CREATE event data to the object.

        Args:
            event: raw message event
        """
        msg = self.cache.place_message_data(event.data)
        if not msg._guild_id and event.data.get("guild_id"):
            msg._guild_id = event.data["guild_id"]
            # todo: Determine why this isn't set *always*

        if not msg.author:
            # sometimes discord will only send an author ID, not the author. this catches that
            msg.channel = (
                await self.cache.get_channel(to_snowflake(msg._channel_id)) if not msg.channel else msg.channel
            )
            if msg._guild_id:
                msg.guild = await self.cache.get_guild(msg._guild_id) if not msg.guild else msg.guild
                msg.author = await self.cache.get_member(msg._guild_id, msg._author_id)
            else:
                msg.author = await self.cache.get_user(to_snowflake(msg._author_id))

        self.dispatch(events.MessageCreate(msg))

    @listen()
    async def _on_raw_message_delete(self, event: RawGatewayEvent) -> None:
        """
        Process raw deletions and dispatch a processed deletion event.
        Args:
            event: raw message deletion event
        """
        message = await self.cache.get_message(
            event.data.get("channel_id"), event.data.get("id"), request_fallback=False
        )
        self.dispatch(events.MessageDelete(message))

    @listen()
    async def _on_raw_message_update(self, event: RawGatewayEvent) -> None:
        """
        Process raw message update event and dispatch a processed update event.

        Args:
            event: raw message update event
        """

        # a copy is made because the cache will update the original object in memory
        before = copy.copy(
            await self.cache.get_message(event.data.get("channel_id"), event.data.get("id"), request_fallback=False)
        )
        after = self.cache.place_message_data(event.data)
        self.dispatch(events.MessageUpdate(before=before, after=after))



import asyncio
import copy
import functools
import logging
import re
from typing import Callable, Coroutine, Any, TYPE_CHECKING

import attr

from dis_snek.const import MISSING, logger_name
from dis_snek.errors import CommandOnCooldown, CommandCheckFailure, MaxConcurrencyReached
from dis_snek.mixins.serialization import DictSerializationMixin
from dis_snek.models.cooldowns import Cooldown, Buckets, MaxConcurrency
from dis_snek.utils.attr_utils import docs
from dis_snek.utils.misc_utils import get_parameters
from dis_snek.utils.serializer import no_export_meta

if TYPE_CHECKING:
    from dis_snek.models.context import Context

log = logging.getLogger(logger_name)

kwargs_reg = re.compile(r"^\*\*\w")
args_reg = re.compile(r"^\*\w")


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class BaseCommand(DictSerializationMixin):
    """
    An object all commands inherit from.
    Outlines the basic structure of a command, and handles checks.

    attributes:
        scale: The scale this command belongs to.
        enabled: Whether this command is enabled
        checks: Any checks that must be run before this command can be run
        callback: The coroutine to be called for this command
        error_callback: The coroutine to be called when an error occurs
        pre_run_callback: A coroutine to be called before this command is run **but** after the checks
        post_run_callback: A coroutine to be called after this command has run

    """

    scale: Any = attr.ib(default=None, metadata=docs("The scale this command belongs to") | no_export_meta)

    enabled: bool = attr.ib(default=True, metadata=docs("Whether this can be run at all") | no_export_meta)
    checks: list = attr.ib(factory=list, metadata=docs("Any checks that must be *checked* before the command can run"))
    cooldown: Cooldown = attr.ib(default=MISSING, metadata=docs("An optional cooldown to apply to the command"))
    max_concurrency: MaxConcurrency = attr.ib(
        default=MISSING, metadata=docs("An optional maximum number of concurrent instances to apply to the command")
    )

    callback: Callable[..., Coroutine] = attr.ib(
        default=None, metadata=docs("The coroutine to be called for this command") | no_export_meta
    )
    error_callback: Callable[..., Coroutine] = attr.ib(
        default=None, metadata=no_export_meta | docs("The coroutine to be called when an error occurs")
    )
    pre_run_callback: Callable[..., Coroutine] = attr.ib(
        default=None,
        metadata=no_export_meta
        | docs("The coroutine to be called before the command is executed, **but** after the checks"),
    )
    post_run_callback: Callable[..., Coroutine] = attr.ib(
        default=None, metadata=no_export_meta | docs("The coroutine to be called after the command has executed")
    )

    def __attrs_post_init__(self):
        if self.callback is not None:
            if hasattr(self.callback, "checks"):
                self.checks += self.callback.checks
            if hasattr(self.callback, "cooldown"):
                self.cooldown = self.callback.cooldown
            if hasattr(self.callback, "max_concurrency"):
                self.max_concurrency = self.callback.max_concurrency

    async def __call__(self, context, *args, **kwargs):
        """
        Calls this command.

        parameters:
            context: The context of this command
            args: Any
            kwargs: Any
        """
        try:
            if await self._can_run(context):
                if self.pre_run_callback is not None:
                    await self.pre_run_callback(context, *args, **kwargs)

                if self.scale is not None and self.scale.scale_prerun:
                    await self.scale.scale_prerun(context, *args, **kwargs)

                await self.call_callback(self.callback, context)

                if self.post_run_callback is not None:
                    await self.post_run_callback(context, *args, **kwargs)

                if self.scale is not None and self.scale.scale_postrun:
                    await self.scale.scale_postrun(context, *args, **kwargs)

        except Exception as e:
            if self.error_callback:
                await self.error_callback(e, context, *args, **kwargs)
            else:
                raise
        finally:
            if self.max_concurrency is not MISSING:
                await self.max_concurrency.release(context)

    async def call_callback(self, callback: Callable, context: "Context"):
        callback = functools.partial(callback, context)  # first param must be ctx
        parameters = get_parameters(callback)
        args = []
        kwargs = {}
        if len(parameters) == 0:
            # if no params, user only wants context
            return await callback()

        c_args = copy.copy(context.args)
        for param in parameters.values():
            if config := getattr(param.annotation, "_annotation_dat", None):
                # if user has used an snek-annotation, run the annotation, and pass the result to the user
                local = {"context": context, "scale": self.scale}
                ano_args = [local[c] for c in config["args"]]
                if param.kind != param.POSITIONAL_ONLY:
                    kwargs[param.name] = param.annotation(*ano_args)
                else:
                    args.append(param.annotation(*ano_args))
                continue
            elif param.name in context.kwargs:
                # if parameter is in kwargs, user obviously wants it, pass it
                if param.kind != param.POSITIONAL_ONLY:
                    kwargs[param.name] = context.kwargs[param.name]
                else:
                    args.append(context.kwargs[param.name])
                if context.kwargs[param.name] in c_args:
                    c_args.remove(context.kwargs[param.name])
            elif param.default is not param.empty:
                kwargs[param.name] = param.default
            else:
                if not str(param).startswith("*"):
                    if param.kind != param.KEYWORD_ONLY:
                        try:
                            args.append(c_args.pop(0))
                        except IndexError:
                            raise ValueError(
                                f"{context.invoked_name} expects {len([p for p in parameters.values() if p.default is p.empty])+len(callback.args)}"
                                f" arguments but received {len(context.args)} instead"
                            ) from None
                    else:
                        raise ValueError(f"Unable to resolve argument: {param.name}")

        if any(kwargs_reg.match(str(param)) for param in parameters.values()):
            # if user has `**kwargs` pass all remaining kwargs
            kwargs = kwargs | {k: v for k, v in context.kwargs.items() if k not in kwargs}
        if any(args_reg.match(str(param)) for param in parameters.values()):
            # user has `*args` pass all remaining args
            args = args + c_args
        return await callback(*args, **kwargs)

    async def _can_run(self, context):
        """
        Determines if this command can be run.

        parameters:
            context: The context of the command
        """
        max_conc_acquired = False  # signals if a semaphore has been acquired, for exception handling

        try:
            if not self.enabled:
                return False

            for _c in self.checks:
                if not await _c(context):
                    raise CommandCheckFailure(self, _c, context)

            if self.scale and self.scale.scale_checks:
                for _c in self.scale.scale_checks:
                    if not await _c(context):
                        raise CommandCheckFailure(self, _c, context)

            if self.max_concurrency is not MISSING:
                if not await self.max_concurrency.acquire(context):
                    raise MaxConcurrencyReached(self, self.max_concurrency)

            if self.cooldown is not MISSING:
                if not await self.cooldown.acquire_token(context):
                    raise CommandOnCooldown(self, await self.cooldown.get_cooldown(context))

            return True

        except Exception:
            if max_conc_acquired:
                await self.max_concurrency.release(context)
            raise

    def error(self, call: Callable[..., Coroutine]):
        """A decorator to declare a coroutine as one that will be run upon an error."""
        if not asyncio.iscoroutinefunction(call):
            raise TypeError("Error handler must be coroutine")
        self.error_callback = call
        return call

    def pre_run(self, call: Callable[..., Coroutine]):
        """A decorator to declare a coroutine as one that will be run before the command"""
        if not asyncio.iscoroutinefunction(call):
            raise TypeError("pre_run must be coroutine")
        self.pre_run_callback = call
        return call

    def post_run(self, call: Callable[..., Coroutine]):
        """A decorator to declare a coroutine as one that will be run after the command has"""
        if not asyncio.iscoroutinefunction(call):
            raise TypeError("post_run must be coroutine")
        self.post_run_callback = call
        return call


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class MessageCommand(BaseCommand):
    """
    Represents a command triggered by standard message.
    """

    name: str = attr.ib(metadata=docs("The name of the command"))


def message_command(
    name: str = None,
):
    """
    A decorator to declare a coroutine as a message command.

    parameters:
        name: The name of the command, defaults to the name of the coroutine
    returns:
        Message Command Object
    """

    def wrapper(func):
        if not asyncio.iscoroutinefunction(func):
            raise ValueError("Commands must be coroutines")
        cmd = MessageCommand(name=name or func.__name__, callback=func)
        return cmd

    return wrapper


def check(check: Callable[..., Coroutine]):
    """
    Add a check to a command.

    parameters:
        check: A coroutine as a check for this command
    """

    def wrapper(coro):
        if isinstance(coro, BaseCommand):
            coro.checks.append(check)
            return
        if not hasattr(coro, "checks"):
            coro.checks = []
        coro.checks.append(check)
        return coro

    return wrapper


def cooldown(bucket: Buckets, rate: int, interval: float):
    """
    Add a cooldown to a command

    Args:
        bucket: The bucket used to track cooldowns
        rate: How many commands may be ran per interval
        interval: How many seconds to wait for a cooldown
    """

    def wrapper(coro: Callable[..., Coroutine]):
        cooldown_obj = Cooldown(bucket, rate, interval)

        coro.cooldown = cooldown_obj

        return coro

    return wrapper


def max_concurrency(bucket: Buckets, concurrent: int):
    """
    Add a maximum number of concurrent instances to the command.

    Args:
        bucket: The bucket to enforce the maximum within
        concurrent: The maximum number of concurrent instances to allow
    """

    def wrapper(coro: Callable[..., Coroutine]):
        max_conc = MaxConcurrency(concurrent, bucket)

        coro.max_concurrency = max_conc

        return coro

    return wrapper



# this could be done programmatically, but it's better practice to manually handle the imports

from .application_commands import *
from .checks import *
from .color import *
from .command import *
from .context import *
from .cooldowns import *
from .discord import *
from .file import *
from .listener import *
from .route import *
from .scale import *
from .snowflake import *
from .timestamp import *

# these are intentionally not imported, leaving them here so people dont think its a mistake
# from .events import *
# from .enums import *


from .discord_objects.activity import *
from .discord_objects.application import *
from .discord_objects.asset import *
from .discord_objects.channel import *
from .discord_objects.components import *
from .discord_objects.embed import *
from .discord_objects.emoji import *
from .discord_objects.guild import *
from .discord_objects.invite import *
from .discord_objects.message import *
from .discord_objects.reaction import *
from .discord_objects.role import *
from .discord_objects.sticker import *
from .discord_objects.team import *
from .discord_objects.thread import *
from .discord_objects.user import *
from .discord_objects.webhooks import *



import asyncio
import logging
from typing import TYPE_CHECKING, Callable, Coroutine

from dis_snek.const import logger_name
from dis_snek.models import events, SnakeBotUser

if TYPE_CHECKING:
    from dis_snek.smart_cache import GlobalCache

log = logging.getLogger(logger_name)


class EventMixinTemplate:
    """All event mixins inherit from this to keep them uniform"""

    cache: "GlobalCache"
    dispatch: Callable[[events.BaseEvent], None]
    _init_interactions: Callable[[], Coroutine]
    synchronise_interactions: Callable[[], Coroutine]
    _user: SnakeBotUser
    _guild_event: asyncio.Event



import logging

from dis_snek.const import logger_name
from dis_snek.event_processors._template import EventMixinTemplate
from dis_snek.models import listen, events, Invite
from dis_snek.models.events import RawGatewayEvent
from dis_snek.utils.converters import timestamp_converter

log = logging.getLogger(logger_name)


class ChannelEvents(EventMixinTemplate):
    @listen()
    async def _on_raw_channel_create(self, event: RawGatewayEvent) -> None:

        channel = self.cache.place_channel_data(event.data)
        self.dispatch(events.ChannelCreate(channel))

    @listen()
    async def _on_raw_channel_delete(self, event: RawGatewayEvent) -> None:
        # for some reason this event returns the deleted object?
        # so we cache it regardless
        channel = self.cache.place_channel_data(event.data)
        self.dispatch(events.ChannelDelete(channel))

    @listen()
    async def _on_raw_channel_update(self, event: RawGatewayEvent) -> None:
        channel = self.cache.place_channel_data(event.data)
        self.dispatch(events.ChannelUpdate(channel))

    @listen()
    async def _on_raw_channel_pins_update(self, event: RawGatewayEvent) -> None:
        channel = await self.cache.get_channel(event.data.get("channel_id"))
        channel.last_pin_timestamp = timestamp_converter(event.data.get("last_pin_timestamp"))
        self.cache.channel_cache[channel.id] = channel
        self.dispatch(events.ChannelPinsUpdate(channel, channel.last_pin_timestamp))

    @listen()
    async def _on_raw_invite_create(self, event: RawGatewayEvent) -> None:
        self.dispatch(events.InviteCreate(Invite.from_dict(event.data, self)))  # type: ignore

    @listen()
    async def _on_raw_invite_delete(self, event: RawGatewayEvent) -> None:
        self.dispatch(events.InviteDelete(Invite.from_dict(event.data, self)))  # type: ignore



from typing import Union, List, Optional, Type

from dis_snek.models import (
    OptionTypes,
    SlashCommandChoice,
    ChannelTypes,
    SlashCommandOption,
    User,
    BaseChannel,
    Role,
    Member,
)


def slash_str_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union[SlashCommandChoice, dict]] = None,
) -> Type[str]:
    """
    Annotates an argument as a string type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
    """
    option = SlashCommandOption(
        name="placeholder",
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices,
        type=OptionTypes.STRING,
    )
    return option  # type: ignore


def slash_float_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union[SlashCommandChoice, dict]] = None,
    min_value: Optional[float] = None,
    max_value: Optional[float] = None,
) -> Type[float]:
    """
    Annotates an argument as a float type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
        min_value:
        max_value:
    """
    option = SlashCommandOption(
        name="placeholder",
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices,
        max_value=max_value,
        min_value=min_value,
        type=OptionTypes.NUMBER,
    )
    return option  # type: ignore


def slash_int_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union[SlashCommandChoice, dict]] = None,
    min_value: Optional[float] = None,
    max_value: Optional[float] = None,
) -> Type[int]:
    """
    Annotates an argument as a integer type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
        min_value:
        max_value:
    """
    option = SlashCommandOption(
        name="placeholder",
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices,
        max_value=max_value,
        min_value=min_value,
        type=OptionTypes.INTEGER,
    )
    return option  # type: ignore


def slash_bool_option(
    description: str,
    required: bool = False,
) -> Type[bool]:
    """
    Annotates an argument as a boolean type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
    """
    option = SlashCommandOption(
        name="placeholder",
        description=description,
        required=required,
        type=OptionTypes.BOOLEAN,
    )
    return option  # type: ignore


def slash_user_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
) -> Type[Union[User, Member]]:
    """
    Annotates an argument as a user type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
    """
    option = SlashCommandOption(
        name="placeholder",
        description=description,
        required=required,
        autocomplete=autocomplete,
        type=OptionTypes.USER,
    )
    return option  # type: ignore


def slash_channel_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union[SlashCommandChoice, dict]] = None,
    channel_types: Optional[list[Union[ChannelTypes, int]]] = None,
) -> Type[BaseChannel]:
    """
    Annotates an argument as a channel type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
        channel_types: The types of channel allowed by this option
    """
    option = SlashCommandOption(
        name="placeholder",
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices,
        channel_types=channel_types,
        type=OptionTypes.CHANNEL,
    )
    return option  # type: ignore


def slash_role_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union[SlashCommandChoice, dict]] = None,
) -> Type[Role]:
    """
    Annotates an argument as a role type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
    """
    option = SlashCommandOption(
        name="placeholder",
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices,
        type=OptionTypes.ROLE,
    )
    return option  # type: ignore


def slash_mentionable_option(
    description: str,
    required: bool = False,
    autocomplete: bool = False,
    choices: List[Union[SlashCommandChoice, dict]] = None,
) -> Type[Union[Role, BaseChannel, User, Member]]:
    """
    Annotates an argument as a mentionable type slash command option.

    Args:
        description: The description of your option
        required: Is this option required?
        autocomplete: Use autocomplete for this option
        choices: The choices allowed by this command
    """
    option = SlashCommandOption(
        name="placeholder",
        description=description,
        required=required,
        autocomplete=autocomplete,
        choices=choices,
        type=OptionTypes.MENTIONABLE,
    )
    return option  # type: ignore



import copy
import logging

from dis_snek.const import logger_name, MISSING
from dis_snek.event_processors._template import EventMixinTemplate
from dis_snek.models import listen, events
from dis_snek.models.events import RawGatewayEvent

log = logging.getLogger(logger_name)


class RoleEvents(EventMixinTemplate):
    @listen()
    async def _on_raw_guild_role_create(self, event: RawGatewayEvent) -> None:
        g_id = event.data.get("guild_id")
        role = self.cache.place_role_data(g_id, [event.data.get("role")])
        self.dispatch(events.RoleCreate(g_id, role[int(event.data["role"]["id"])]))

    @listen()
    async def _on_raw_guild_role_update(self, event: RawGatewayEvent) -> None:
        g_id = event.data.get("guild_id")
        r_data = event.data.get("role")
        before = copy.copy(await self.cache.get_role(g_id, r_data["id"], False) or MISSING)

        after = self.cache.place_role_data(g_id, [r_data])
        after = after[int(event.data["role"]["id"])]

        self.dispatch(events.RoleUpdate(g_id, before, after))

    @listen()
    async def _on_raw_guild_role_delete(self, event: RawGatewayEvent) -> None:
        self.dispatch(events.RoleDelete(event.data.get("guild_id"), event.data.get("role_id")))



from typing import Union, TYPE_CHECKING

from dis_snek.models.discord_objects.role import Role
from dis_snek.models.snowflake import Snowflake_Type, to_snowflake
from dis_snek.models.context import Context


def has_role(role: Union[Snowflake_Type, Role]):
    """
    Check if the user has the given role

    Args:
        role: The Role or role id to check for
    """

    async def check(ctx: Context) -> bool:
        if ctx.guild is None:
            return False
        return await ctx.author.has_role(role)

    return check


def has_any_role(*roles: Union[Snowflake_Type, Role]):
    """
    Checks if the user has any of the given roles
    Args:
        *roles: The Role(s) or role id(s) to check for
    """

    async def check(ctx: Context) -> bool:
        if ctx.guild is None:
            return False

        if any(ctx.author.has_role(to_snowflake(r)) for r in roles):
            return True
        return False


def has_id(user_id):
    """
    Checks if the author has the desired ID.

    parameters:
        coro: the function to check
    """

    async def check(ctx: Context) -> bool:
        return ctx.author.id == user_id

    return check


def is_owner():
    """
    Is the author the owner of the bot.

    !!! note
        this does not account for teams
    parameters:
        coro: the function to check
    """

    async def check(ctx: Context) -> bool:
        return ctx.author.id == ctx.bot.owner

    return check


def guild_only():
    """
    This command may only be ran in a guild
    """

    async def check(ctx: Context) -> bool:
        return ctx.guild is not None

    return check


def dm_only():
    """
    This command may only be ran in a dm
    """

    async def check(ctx: Context) -> bool:
        return ctx.guild is None

    return check



import logging

from dis_snek.const import logger_name
from dis_snek.event_processors._template import EventMixinTemplate
from dis_snek.models import listen, events, StageInstance
from dis_snek.models.events import RawGatewayEvent

log = logging.getLogger(logger_name)


class StageEvents(EventMixinTemplate):
    @listen()
    async def _on_raw_stage_instance_create(self, event: RawGatewayEvent) -> None:
        self.dispatch(events.StageInstanceCreate(StageInstance.from_dict(event.data, self)))  # type: ignore

    @listen()
    async def _on_raw_stage_instance_update(self, event: RawGatewayEvent) -> None:
        self.dispatch(events.StageInstanceUpdate(StageInstance.from_dict(event.data, self)))  # type: ignore

    @listen()
    async def _on_raw_stage_instance_delete(self, event: RawGatewayEvent) -> None:
        self.dispatch(events.StageInstanceDelete(StageInstance.from_dict(event.data, self)))  # type: ignore
